/*!
 * Jodit Editor (https://xdsoft.net/jodit/)
 * License https://xdsoft.net/jodit/license.html
 * Copyright 2013-2018 Valeriy Chupurnov https://xdsoft.net
 */

import {Jodit} from "../Jodit";
import {css, debounce, dom, throttle} from "../modules/Helpers";
import {Config} from '../Config'

declare module "../Config" {
    interface Config {
        allowResizeX: boolean;
        allowResizeY: boolean;
    }
}

Config.prototype.allowResizeX = false;
Config.prototype.allowResizeY = true;

/**
 * Resize editor
 * @param {Jodit} editor
 */
export function size(editor: Jodit) {
    const setHeight = (height: number | string) => {
        css(editor.container, 'height', height);
        if (editor.options.saveHeightInStorage) {
            editor.storage.set('height', height);
        }
    };

    const setWidth = (width: number | string) => css(editor.container, 'width', width);
    const setHeightWorkPlace = (height: number | string) => css(editor.workplace, 'height', height);
    // const setWidthWorkPlace = (width: number | string) => css(editor.workplace, 'width', width);


    if (editor.options.height !== 'auto' && (editor.options.allowResizeX || editor.options.allowResizeY)) {
        const handle: HTMLAnchorElement = <HTMLAnchorElement>dom('<div class="jodit_editor_resize" ><a href="javascript:void(0)"></a></div>', editor.ownerDocument),
            start: { x: number, y: number, w: number, h: number } = {
                x: 0, y: 0, w: 0, h: 0
            };

        let isResized: boolean = false;

        editor.events
            .on(handle, 'mousedown touchstart', (e: MouseEvent) => {
                isResized = true;
                start.x = e.clientX;
                start.y = e.clientY;
                start.w = editor.container.offsetWidth;
                start.h = editor.container.offsetHeight;
                editor.lock();
                e.preventDefault();
            })
            .on(editor.ownerWindow, 'mousemove touchmove', throttle((e: MouseEvent) => {
                if (isResized) {
                    setHeight(editor.options.allowResizeY ? start.h + e.clientY - start.y: start.h);
                    setWidth(editor.options.allowResizeX ? start.w + e.clientX - start.x : start.w);
                    resizeWorkspaceImd();
                    editor.events.fire('resize');
                }
            }, editor.options.observer.timeout / 10))
            .on(editor.ownerWindow, 'mouseup touchsend', () => {
                if (isResized) {
                    isResized = false;
                    editor.unlock();
                }
            })
            .on('afterInit', () => {
                editor.container.appendChild(handle);
            })
            .on('toggleFullSize', (fullsize: boolean) => {
                handle.style.display = fullsize ? 'none' : 'block';
            });
    }

    const getNotWorkHeight = (): number => (editor.options.toolbar ? editor.toolbar.container.offsetHeight : 0) + (editor.statusbar ? editor.statusbar.container.offsetHeight : 0);;
    const calcMinHeightWorkspace = () => {
        if (!editor.container || !editor.container.parentNode) {
            return;
        }

        const minHeight: number = <number>css(editor.container, 'minHeight') - getNotWorkHeight();

        [editor.workplace, editor.iframe, editor.editor].map(elm => {
            let minHeightD : number = elm === editor.editor ? minHeight - 2 : minHeight; // borders
            elm && css(<HTMLElement>elm, 'minHeight', minHeightD)
        });
    };

    const resizeWorkspaceImd = () => {
        calcMinHeightWorkspace();

        if (editor.container && (editor.options.height !== 'auto' || editor.isFullSize())) {
            setHeightWorkPlace(editor.container.offsetHeight - getNotWorkHeight())
        }
    };

    const resizeWorkspace = debounce(resizeWorkspaceImd, editor.options.observer.timeout);

    editor.events
        .on('toggleFullSize', (fullsize: boolean) => {
            if (!fullsize && editor.options.height === 'auto') {
                setHeightWorkPlace('auto');
                calcMinHeightWorkspace();
            }
        })
        .on('afterInit', () => {
            css(editor.editor, {
                minHeight: '100%'
            });
            css(editor.container, {
                minHeight: editor.options.minHeight
            });

            let height: string | number = editor.options.height;

            if (editor.options.saveHeightInStorage && height !== 'auto') {
                let localHeight: string | null = editor.storage.get('height');
                if (localHeight) {
                    height = localHeight
                }
            }

            setHeight(height);
            setWidth(editor.options.width);

            resizeWorkspaceImd();
        }, undefined, undefined,true)
        .on(window, 'load', resizeWorkspace)
        .on('afterInit resize updateToolbar scroll afterResize', resizeWorkspace)

}